module EventObject
  refine Array do
    # Emit event with specified arguments and specified block.
    #
    # @example
    #    using EventObject
    #    event = [() -> 1, () -> 2]
    #    p event.fire
    #    # [1, 2]
    #
    # @param *a [...] arguments to be passed
    # @param &b [Block] block to be passed
    # @return [Array] the results of the call
    def fire(*a, &b)
      dup.map { |h| h.call(*a, &b) }
    end
    alias emit fire

    # Emit event with specified arguments within the context. It's giving the each listener
    # access to object intance variables. Unlike fire, the block is not allowed.
    #
    # @param context [Object] the context of the receiver
    # @return [Array]
    def instance_fire(context, *a)
      dup.map { |h| context.instance_exec(*a, &h) }
    end

    # Removes the specified listener from the array. If no arguments are supplied then
    # all listeners will be removed.
    #
    # @example
    #   events = [() -> 1, () -> 2]
    #   events.off
    #
    # @param listener [Proc] listener or nil
    # @return [Proc, Array] listener or nil or array of listeners
    def off(listener=nil)
      listener ? delete(listener) : clear
    end
  end
end
