module EventObject
  refine Object do

    # Add the `listener` block for the specified `event`. `on` calls can be chained.
    #
    # @example
    #    class EventEmitter
    #      using Emmy
    #      events :success, :error
    #    end
    #    emitter = EventEmitter.new
    #    emitter.on :success do
    #      [200, {}, "OK"]
    #    end
    #    p emitter.success!
    #
    # @param event [Symbol]
    # @param listener [Proc]
    # @param block [Block]
    # @return [Proc]
    def on(event, listener=nil, &block)
      var(event, Array).tap { |e| e << (listener || block) }; (listener || block)
    end

    # Add the `listener` block for the specified `event`. When event happened `listener` will be removed.
    #
    # @param event
    # @param listener
    # @param block
    # @return [Proc]
    def once(event, listener=nil, &block)
      (ev = nil).tap { |ev|
        var(event, Array).tap { |e| e << (ev = ->(*a) { (listener || block).call(*a); e.delete(ev) }) }
      }
    end

    # Remove events depending on the argument counts.
    #
    # @example
    #   obj.off(:event)             # Remove all event listensers for the specified `event`
    #   obj.off(:event, proc)       # Remove an event listener
    #   obj.off(:ev, object, :meth)
    # @param args [...]
    # @return []
    def off(*args)
      case args.size
      when 1 then var(args[0], Array).clear
      when 2 then var(args[0], Array).delete(args[1])
      when 3 then var(args[0], Array).delete(args[1].method(args[2]))
      else
        raise "too many arguments"
      end
      self
    end

    # Add the `instance` method with `name` like block for the specified event.
    #
    # @param instance [Object]
    # @param event [Symbol]
    # @param name [Symbol]
    # @return []
    def listen(instance, event, name)
      instance.on event, method(name)
    end

    # Remove all listeners being methods of the `instance`.
    #
    # @param []
    # @return []
    def stop_listen(instance, event)
      instance.var(event, Array).delete_if { |e| e.is_a?(Method) && e.receiver == self }
    end

    # Define an instance method. The method will be appeared only in this instance.
    #
    # @param symbol [Symbol]
    # @return [Symbol]
    def def(symbol, &b)
      define_singleton_method symbol, &b
    end

    # Get an instance variable. If the variable is undefined then the variable will be initialized.
    #
    # @param symbol [Symbol]
    # @param init [...]
    # @return [Symbol]
    def var(symbol, init=nil)
      instance_variable_get(sym = "@#{symbol}") || instance_variable_set(sym, (init.is_a?(Class) ? init.new : init))
    end
  end
end
